/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

// This parser produces a blurp on the output device once a second. It does not
// take any input. It is useful for diagnostics.

#include "data_structs.h"
#include "ParserRegistration.h"
#include "sched.h"
#include "MIDIsvc.h"
#include "MIDI_lib.h"


static Bool running;
static ioref ou_node;
static scheduler *schedp;
static MIDIsvc *oap;

class BlurpParser : public parser
{
public:
	BlurpParser(Rogus_McBogus*);
	~BlurpParser() { running = False; }
};

static ScheduleTime func(void*)
{
	if (!running)
		return NEVER;
	ScheduleTime t = oap->time();
	int i;
	for (i = 0; i < 13; i++)
	{
		MIDI_msg m;
		m.set(MIDI_msg::note_on, 0, uchar(60+i), 64, ou_node);
		task_data td(m, 500, t+(50*(i+1)));
		schedp->insert(td);
	}
	return 2000;
}

BlurpParser::
BlurpParser(Rogus_McBogus *r)
{
   schedp = &r->s;
   oap = &r->o;

	ou_node = oap->open_output(); // open the default output
	running = True;
	task_data td(func, NULL, oap->time()+2000);
	schedp->insert(td);
}

static parser* newFunc(Rogus_McBogus *r)
{
	return (new BlurpParser(r));
}

static ParserRegistration reg(newFunc, "BlurpParser");

// Author: Patrick Pelletier
