/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

// This parser will send sysex to a K2500 to read its screen contents,
// and will update the screen contents whenever a button is pressed.

#include "ParserRegistration.h"
#include "Rogus_McBogus.h"


class K2500ScreenParser : public parser
{
public:
	K2500ScreenParser(Rogus_McBogus*);
	~K2500ScreenParser();
};

static Bool running;
static ioref ou_node;
static Rogus_McBogus* r;

static ScheduleTime func(void*)
{
	if (!running)
		return NEVER;
	unsigned char message[] = { 0xf0, // start of sysex
								0x07, // Kurzweil Manufacturer ID
								0x00, // Device ID
								0x78, // Product Identifier
								0x15, // ALLTEXT message type
								0xf7}; // end of sysex
	MIDI_msg m;
	m.node = ou_node;
	m.set_sys_ex(message, 6);
	r->o.write(&m);
	return NEVER;
}

static void infunc(MIDI_msg* m)
{
	if (m->ed == NULL)
		return; // ignore non-sysex
	if (m->ed->st != smsgt::sys_ex)
	{
		r->ew.w("sanity check failed");
		return;
	}
	int l = m->ed->sys_ex.l;
	uchar* d = m->ed->sys_ex.d;
	if (l < 6)
	{
		r->tw.pr("got a really short sysex message\n");
		return;
	}
	if (d[1] != 7 || d[3] != 120)
	{
		r->tw.pr("got sysex not from K2500\n");
		return;
	}
	if (d[4] == 0x14 && (d[5] == 0x08 || d[5] == 0x0d)) // if button release, poll screen soon
	{
		task_data td(func, NULL, r->o.time()+800);
		r->s.insert(td);
	}
	else if (d[4] == 0x19) // screen reply
	{
		if (l != 327)
		{
			r->tw.pr("got %d bytes instead of 327 bytes\n",l);
			return;
		}
		int i;
		r->tw.pr("\n");
		for (i = 0; i < 8; i++)
		{
			int j;
			for (j = 0; j < 40; j++)
				r->tw.pr("%c",d[5+j+i*40]);
			r->tw.pr("\n");
		}
	}
}

K2500ScreenParser::
K2500ScreenParser(Rogus_McBogus *in_r)
{
   r = in_r;
	ou_node = r->o.open_output(); // open the default output
	r->o.open_input(True); // open the default input
	msg_hnd[MIDI_msg::system] = infunc;
	running = True;
	task_data td(func, NULL, r->o.time()+800);
	r->s.insert(td);
}

K2500ScreenParser::
~K2500ScreenParser()
{
	running = False;
}

static parser* newFunc(Rogus_McBogus *r)
{
	return (new K2500ScreenParser(r));
}

static ParserRegistration reg(newFunc, "K2500ScreenParser");

// Author: Patrick Pelletier
