/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

// This parser is an example of how to use the Rogus score player. The MIDI file
// "test_MIDI_file.mid" from the current directory is read in, and each key on the
// input device is mapped to play this score at a different speed and transposition.

#include <math.h>
#include "Rogus_McBogus.h"
#include "ParserRegistration.h"
#include "Transposer.h"
#include "ScorePlayer.h"
#include "Score.h"
#include "ScorePlayer.h"

static scheduler *schedp;
static MIDIsvc* oap;
static Rogus_McBogus *rogus;

class score_parser : public parser
{
	typedef UtilPtr<Score> ScorePtr;
public:
	score_parser();
	~score_parser();
};

static ScorePlayer* sp_array[128];

static
void
note_on( MIDI_msg *m ) 
{
	if (m->d2 == 0)
		sp_array[m->d1]->stop();
	else
		sp_array[m->d1]->start();
}

static
void
note_off( MIDI_msg *m )
{
	sp_array[m->d1]->stop();
}

score_parser::
score_parser()
{
   Filedata f("test_MIDI_file.mid",&rogus->ew);
   ScorePtr myScore = new Score(rogus,&f);
	oap->open_input(); // open the default input
	
	for (int i = 0; i < 128; i++)
	{
		sp_array[i] = new ScorePlayer(rogus,myScore);
		sp_array[i]->addFilter(new Transposer(i-60));
		sp_array[i]->setScale((ScoreScale)(pow(double(2), (((double)(60-i))/12))));
		sp_array[i]->setLoop(1);
	}

	msg_hnd[MIDI_msg::note_on] = note_on;
	msg_hnd[MIDI_msg::note_off] = note_off;
}

score_parser::
~score_parser()
{
	int i;
	for (i = 0; i < 125; i++)
	{
		sp_array[i]->stop();
		delete sp_array[i];
	}
	for (i = 125; i < 128; i++)
	{
		sp_array[i]->stop();
		delete sp_array[i];
	}
}

static parser* newFunc(Rogus_McBogus *r)
{
   rogus = r;
   schedp = &r->s;
   oap = &r->o;
	return (new score_parser());
}

static ParserRegistration reg(newFunc, "score_parser");

// Authors: Ben Denckla, Patrick Pelletier
