/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#ifndef MIDISVC_H
#define MIDISVC_H

#pragma warning(disable : 4512)
// this is the warning for "<class>: assignment operator could not be generated"

#include "err_and_warn.h"
#include "basic_types.h"
#include "BinStatistic.h"
#include "Midiator.h"
#include "WrappedCS.h"
#include "ctype.h" // for toupper()
#include "data_structs.h"
#include "MIDI_lib.h"
#include "Rogus.h"

enum { my_sysex_input_size = 128, my_sysex_num_buffers = 50 };

struct MySysexInputBuffer {
	MIDIHDR hdr;
	unsigned char buf[my_sysex_input_size];
};

#pragma warning(disable: 4097)
  /*
      4097 causes the following two warnings:
      "typedef-name 'MIDIINCAPS' used as synonym for class-name 'tagMIDIINCAPSA'"
      "typedef-name 'MIDIOUTCAPS' used as synonym for class-name 'tagMIDIOUTCAPSA'"
  */

class MyMidiIn : public MIDIINCAPS {
public:
   enum type { NORMAL, MIDIATOR };

   Bool open;
   type ty;

	HMIDIIN handle;
	Midiator* midiator;

   MySysexInputBuffer* sysex;
	MIDIHDR* assembly_order[my_sysex_num_buffers];
	int assembly_position;

   Bool operator==(const MyMidiIn& m) { return (this == &m); }
   MyMidiIn() { open = False; }
   const char *name() { return szPname; }
};

class MyMidiOut : public MIDIOUTCAPS {
public:
   enum type { NORMAL, MIDIATOR };

  	Bool open;
   type ty;

	HMIDIOUT handle;
	Midiator* midiator;

   uchar running_status;
	long running_status_time;

   Bool operator==(const MyMidiOut& m) { return (this == &m); }
   MyMidiOut() { open = False; }
   const char *name() { return szPname; }
};

#pragma warning(default: 4097)

struct MyQueueElement {
	DWORD dwParam1;
	DWORD dwParam2;
	UINT n;
	MIDIHDR* hdr;
};

class scheduler;  // for use below in MIDIsvc's schedp
class user_prefs; // for used below in MIDIsvc's upp

class MIDIsvc
{
public:
	typedef UtilArray<MyMidiIn>       MyMidiInArray;
	typedef UtilPtr  <MyMidiInArray>  MyMidiInArrayPtr;
	typedef UtilArray<MyMidiOut>      MyMidiOutArray;
	typedef UtilPtr  <MyMidiOutArray> MyMidiOutArrayPtr;
private:
   const user_prefs *upp;

	static DWORD our_start_time; // 3
	static MyMidiInArrayPtr inlh;
	static MyMidiOutArrayPtr onlh;
	static c_queue<MyQueueElement> cq;
	BinStatistic queueStat;

	static void CALLBACK my_read_hook(HMIDIIN hMidiIn, UINT wMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2);
	static void my_cdecl_read_hook(HMIDIIN hMidiIn, UINT wMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2);
	
	void init_timer();
	MyMidiInArrayPtr get_input_nodes();
	MyMidiOutArrayPtr get_output_nodes();
	void open_sysex(MyMidiIn *cn, ioref i);

	typedef UtilPtr<Midiator> MidiatorPtr;
	typedef UtilArray<MidiatorPtr> MidiatorPtrArray;
	typedef UtilPtr<MidiatorPtrArray> MidiatorPtrArrayPtr;
	MidiatorPtrArrayPtr midiator_array;
	
	void write_ed(const MidiMsg *m);

	void process_sysex_queue_el(MyQueueElement *q);

	long midiInStart_time_offset;
	static long adjustment;

   static scheduler *schedp;
   static WrappedCS *csp;
   static err_and_warn* ewp;

   static Bool handles(parser *lp, DWORD p)
      { return (lp->msg_hnd[ MidiMsg::msgt(uchar(p & 0xff)) ] != NULL); }

   class GenericNodeList
   {
      virtual const char* name(ioref i)=0;
      virtual ioref length()=0;
     	static void strtoupper( const char *i, char *o )
         { while ( (*o++ = char(toupper(*i++))) != 0 ) {} }
   public:
      err_ind select_by_name( const char*, ioref*, err_and_warn* );
   };

   class InputNodeList : public GenericNodeList
   {
      MyMidiInArrayPtr  nlh;
      const char* name(ioref i) { return (*nlh)[i].name(); }
      ioref length() { return ioref((*nlh).length); }
   public:
      InputNodeList(MyMidiInArrayPtr  in_nlh) { nlh = in_nlh; }
   };

   class OutputNodeList : public GenericNodeList
   {
      MyMidiOutArrayPtr  nlh;
      const char* name(ioref i) { return (*nlh)[i].name(); }
      ioref length() { return ioref((*nlh).length); }
   public:
      OutputNodeList(MyMidiOutArrayPtr  in_nlh) { nlh = in_nlh; }
   };


public:
	static parser *lp;

	 MIDIsvc(const user_prefs*, scheduler*, WrappedCS*, err_and_warn*);
	~MIDIsvc();

	ioref open_input( const char*, Bool allow_sysex=False); // 1
	ioref open_output( const char* ); // renamed for symmetry 5/25/96 PP

	// added these 4 for more generality 5/26/96 PP
	ioref open_input(ioref n, Bool allow_sysex=False);
	ioref open_output(ioref n);
	ioref open_input(Bool allow_sysex=False);
	ioref open_output();

	// for compatibility 5/25/96 PP
	ioref output_refnum( const char* s ) { return open_output(s); }
	
	void closeAllInputs();
	void closeAllOutputs();
	void processQueue(); // called by scheduler

	typedef UtilArray<UtilStringPtr> UtilStringPtrArray;
	typedef UtilPtr<UtilStringPtrArray> UtilStringPtrArrayPtr;

	ioref howManyOutputNodes(); // added this for PerChannel to use. 10/27/95 PP
	ioref howManyInputNodes(); // added for symmetry 5/25/96 PP

	const char* getOutputNodeName(ioref); // added this for Recorder to use 11/7/95 PP
	const char* getInputNodeName(ioref); // added for symmetry 5/25/96 PP

	static long time();

	long get_adj_midiInStart_time_offset() const { return midiInStart_time_offset + adjustment; }

	err_ind setTimecodeInput(ioref, Bool); // added 6/6/96 PP
	err_ind unsetTimecodeInput(); // added 6/6/96 PP

	void write(const MIDI_msg *m);

  	void RogMidiMsgToWin( const MidiMsg& m, DWORD* p, UINT* n );
	void WinMidiMsgToRog( DWORD p1, DWORD p2, UINT n, MidiMsg *m );
};
/*
1. The
existence of the allow_sysex parameter, as well as its somewhat puzzling
default of False, has to do with a precise moment in history when we were using an
old version of the 8Port/SE drivers that barfed badly when asked to
support SysEx.  It doesn't seem too harmful (and the default of False
saves memory by not allocating SysEx input buffers), so, as I've opted to
maintain backward compatibility by leaving it this way.

3. This point of this is to get around the wrapping problem if Windows is run for over 49 days.
By subtracting the start time of our program, we get around this problem, although we still will
have trouble if our program is run for over ~12 days, since ScheduleTime is signed (not unsigned)
and is in units of demimilliseconds, not milliseconds.
*/

#endif
// Authors: Ben Denckla, Patrick Pelletier
