/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#ifndef SCORE_H
#define SCORE_H

#include "Filedata.h"
#include "UtilPtr.h"
#include "PerChannel.h"
#include "MIDI_lib.h"
#include "Recorder.h"
#include "Rogus_McBogus.h"

class ScoreCreationInfo;
class ScorePlayer;

typedef long ScoreTime;

class Score {
public:
	typedef PerChannel<Bool> ChannelUseMap;
	typedef UtilPtr<ChannelUseMap> ChannelUseMapPtr;
private:
	friend class ScorePlayer;
	typedef void (*ScoreFunc)(void*, ScorePlayer*);
	typedef UtilArray<ScoreFunc> ScoreFuncArray;
	typedef UtilPtr<ScoreFuncArray> ScoreFuncArrayPtr;
	typedef void* VoidStar;
	typedef UtilArray<VoidStar> VoidStarArray;
	typedef UtilPtr<VoidStarArray> VoidStarArrayPtr;
	typedef UtilArray<ScoreTime> ScoreTimeArray;
	typedef UtilPtr<ScoreTimeArray> ScoreTimeArrayPtr;
	MidiMsgArrayPtr msgs;
	ChannelUseMapPtr u;
	ScoreFuncArrayPtr funcs;
	VoidStarArrayPtr args;
	ScoreTimeArrayPtr times;
	void read(const Filedata*);
	static void convertTime(MidiMsgArrayPtr, long, ScoreTimeArrayPtr);
	static void convertFuncTime(ScoreTimeArrayPtr stap, ScoreTime lastTime, long lastTick, long thisTick, int* func_index, long currentTempo, long division);
   Rogus_McBogus *r;
public:
	Score(Rogus_McBogus* in_r) { r = in_r; msgs = new MidiMsgArray(0); }
	Score(Rogus_McBogus* in_r, const Filedata* f) { r = in_r; read(f); }
	Score(const ScoreCreationInfo*);
	Score(Rogus_McBogus*, const Recorder*);
	int getNumEvents() const { return msgs->length; }
	ScoreTime getLength() const;
	ChannelUseMapPtr getChannels() const;
	const MIDI_msg& operator[](int i) const { return (*(((Score*)this)->msgs))[i]; }
};

#endif
// Author: Patrick Pelletier
