/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#ifndef SCHED_H
#define SCHED_H

#include "data_structs.h"
#include "MIDIsvc.h"
#include "MIDI_lib.h"
#include "BinStatistic.h"
#include "user_prefs.h"
#include "WrappedCS.h"

typedef long ScheduleTime;

typedef long (*my_wake_funcp)( void* );

class task_data
{
public:
	enum { fa_t, m_t, on_t } ty;
	ScheduleTime nwu; // 1
	// I had to de-unionify these PP 12/17/95
	struct { my_wake_funcp func; void *arg; } fa;
	MIDI_msg m;
	//unsigned long task_num; // used only in p_queue


	task_data() {}

	task_data( my_wake_funcp f, void *a, ScheduleTime t ) { ty=fa_t; fa.func=f; fa.arg=a; nwu=t; }

	task_data( const MIDI_msg& in_m, ScheduleTime t ) { ty = m_t; m = in_m; nwu=t; }

	task_data( const MIDI_msg& on, ScheduleTime dur, ScheduleTime t ) { ty = on_t; m = on; m.time = dur; nwu=t; }
};
/*
1. nwu is my copy of nextWakeUp.  A copy is necessary because when the WakeUpTask is called,
nextWakeUp is set to NEVER by OMS, and we need to use nextWakeup in the WakeUpTask.
Note 1/25/97: I don't know if this comment is still relevant.  Obviously its reference to OMS
is inaccurate, but beyond that I do not know.
*/


//--------------------------------------------------------------------------------------------------

struct task
{
	MMRESULT ott;
	task* chain;
	task_data td;
};

class scheduler
{
public:
   class task_pool : public pool<task>
   {
	   BinStatistic lateness;
	   static int binner(int i) { return (i*5); }
   public:
	   const int size;
	   len_monitor mon;

	    task_pool(int s) : mon( s, "task pool" ), size(s), pool<task>(s, "" ),
          lateness(12, binner, "Scheduler Lateness") { }

	   task* store( const task_data& );

	   void record_lateness( long l ) { lateness.throwInBin((int)l); }
   };

private:
   task_pool tp;
   task* currentTask;
   task* doMeNext;
   task** doMeNext_prev;
   Bool zappedYourTask;
   HANDLE threadHandle;
   UINT threadID;
   HANDLE eventHandle;
   WrappedCS *csp;
   Bool running;
   user_prefs *upp;
   MIDIsvc *oap;
   err_and_warn *ewp;

	static unsigned WINAPI threadMain(LPVOID lpvThreadParm);
	void reorderTasks();
	void waitAndHandle();
	void removeAndExecute();
	err_ind do_insert(task* t);

	void print_task_list(); // not currently used (was used in debugging)
   void scheduler::cr( task_pool::rrv retval );

public:
	scheduler(user_prefs*, WrappedCS*, err_and_warn*);
	~scheduler();

   void set_oap(MIDIsvc *o) { oap = o; }

	err_ind play_note( const MIDI_msg& on, ScheduleTime dur );
	
	err_ind insert( const task_data& td )
	{
		csp->enter();
		task *t = tp.store( td );
		err_ind result = err;
		if (t != NULL)
			result = do_insert(t);
		csp->leave();
		return result;
	}

	// this version is like above but the user can get the task back so they can cancel it if they want
	err_ind insert( const task_data& td, task** task_return )
	{
		csp->enter();
		task *t = tp.store( td );
		*task_return = t;
		err_ind result = err;
		if (t != NULL)
			result = do_insert(t);
		csp->leave();
		return result;
	}

	err_ind cancel(task* t);

	void processQueue() { SetEvent(eventHandle); }
};

#endif
// Authors: Ben Denckla, Patrick Pelletier
