/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include <math.h>
#include "MIDIsvc.h"
#include "Rogus.h"
#include "sched.h"
#include "user_prefs.h"
#include "Rogus_McBogus.h" // for rx

//--------------------------------------------------------------------------------------------------

DWORD MIDIsvc::our_start_time = 0;
static ioref default_output_ioref = 0;
static ioref default_input_ioref = 0;
static HANDLE eventHandle;

long MIDIsvc::adjustment = 0;
static long prev_time;
static Bool skip_forward;
static Bool skip_backward;
static long last_mtc;
static long last_mtc_begin;
static long doubleframes_since_mtc_begin;
static int fps;
static HMIDIIN timecode_input;
static Bool eat_timecode;

scheduler* MIDIsvc::schedp;
WrappedCS* MIDIsvc::csp;
err_and_warn* MIDIsvc::ewp;

MIDIsvc::MyMidiInArrayPtr MIDIsvc::inlh;
MIDIsvc::MyMidiOutArrayPtr MIDIsvc::onlh;


static void got_mtc()
{
//	tw.pr("got mtc\n");
	long now = MIDIsvc::time();
	if (now - last_mtc < (6000/fps) && last_mtc_begin != NEVER) // ignore big delays
	{
		doubleframes_since_mtc_begin++;
		if (now - last_mtc_begin < doubleframes_since_mtc_begin*(4000/fps))
		{
//			tw.pr("we need to set clock ahead 1 dms\n");
			skip_forward = True;
			skip_backward = False;
		}
		else if (now - last_mtc_begin > doubleframes_since_mtc_begin*(4000/fps))
		{
//			tw.pr("we need to set clock back 1 dms\n");
			skip_forward = False;
			skip_backward = True;
		}
	}
	else
	{
		last_mtc_begin = now;
		doubleframes_since_mtc_begin = 0;
	}
	last_mtc = now;
}

long MIDIsvc::time()
{
	long now = (timeGetTime()-our_start_time) * 2;
	if (now > prev_time && skip_forward)
	{
		//tw.pr("setting clock ahead 1 dms\n");
		adjustment++;
		skip_forward = False;
	}
	else if (now > prev_time && skip_backward)
	{
		//tw.pr("setting clock back 1 dms\n");
		adjustment--;
		skip_backward = False;
	}
	prev_time = now;
	return (now + adjustment);
}

err_ind MIDIsvc::setTimecodeInput(ioref i, Bool eat)
{
	if (i >= inlh->length)
	{
		ewp->w("invalid input node %u",i);
		return err;
	}
	if ((*inlh)[i].open == False)
	{
		ewp->w("on \"%s\" (node %d) set timecode input when not open",(*inlh)[i].szPname,i);
		return err;
	}
	timecode_input = (*inlh)[i].handle;
	eat_timecode = eat;
	last_mtc_begin = NEVER;
	return no_err;
}

err_ind MIDIsvc::unsetTimecodeInput()
{
	timecode_input = NULL;
	return no_err;
}

void
MIDIsvc::
init_timer()
{
	if (timeBeginPeriod(1) != TIMERR_NOERROR) // set timer resolution to 1 millisecond
		ewp->w("Cannot set timer resolution to 1 millisecond");
	our_start_time = timeGetTime();
	adjustment = 0;
	prev_time = 0;
	skip_forward = False;
	skip_backward = False;
	last_mtc = 0;
	fps = 25;
}

c_queue<MyQueueElement> MIDIsvc::cq( 1000, "" ); // changed name from "input queue" to empty string 11/12/95 PP
parser *MIDIsvc::lp = 0;

static int binner(int n)
{
	return (int)(pow(2,n));
}

MIDIsvc::MyMidiInArrayPtr MIDIsvc::get_input_nodes()
{
	UINT n = midiInGetNumDevs();

	int totdevs = n + (midiator_array->length * Midiator::numInputs());

	MyMidiInArrayPtr a = new MyMidiInArray(totdevs);

	UINT i;
	for (i = 0; i < n; i++)
	{
      (*a)[i].ty = MyMidiIn::NORMAL;
		if (midiInGetDevCaps(i, &((*a)[i]), sizeof(MIDIINCAPS)) != MMSYSERR_NOERROR)
		{
			ewp->w("Unable to get info for input device %d", i);
			strcpy((*a)[i].szPname, "Unknown");
		}
	}
	UINT j;
	for (i = n, j = 0; j < midiator_array->length; j++)
	{
      (*a)[i].ty = MyMidiIn::MIDIATOR;
		(*midiator_array)[j]->in_start = i;
		UINT k;
		for (k = 0; k < Midiator::numInputs(); k++, i++)
		{
			(*midiator_array)[j]->getInputName(k, ((*a)[i].szPname));
			(*a)[i].midiator = (*midiator_array)[j];
		}
	}

	return a;
}

MIDIsvc::MyMidiOutArrayPtr MIDIsvc::get_output_nodes()
{
	UINT n = midiOutGetNumDevs();

	int totdevs = n + (midiator_array->length * Midiator::numOutputs());

	MyMidiOutArrayPtr a = new MyMidiOutArray(totdevs);

	UINT i;
	for (i = 0; i < n; i++)
	{
      (*a)[i].ty = MyMidiOut::NORMAL;
		if (midiOutGetDevCaps(i, &((*a)[i]), sizeof(MIDIOUTCAPS)) != MMSYSERR_NOERROR)
		{
			ewp->w("Unable to get info for output device %d", i);
			strcpy((*a)[i].szPname, "Unknown");
		}
	}
	UINT j;
	for (i = n, j = 0; j < midiator_array->length; j++)
	{
      (*a)[i].ty = MyMidiOut::MIDIATOR;
		(*midiator_array)[j]->out_start = i;
		UINT k;
		for (k = 0; k < Midiator::numOutputs(); k++, i++)
		{
			(*midiator_array)[j]->getOutputName(k, ((*a)[i].szPname));
			(*a)[i].midiator = (*midiator_array)[j];
		}
	}

	return a;
}

MIDIsvc::
MIDIsvc(const user_prefs *in_upp, scheduler *in_schedp, WrappedCS* in_csp, err_and_warn *in_ewp) :
   upp(in_upp), queueStat(10, binner, "Input Queue Usage")
{
   ewp = in_ewp;
   schedp = in_schedp;
   csp = in_csp;
	init_timer();

	eventHandle = CreateEvent(NULL, FALSE, FALSE, NULL);
	
	midiator_array = new MidiatorPtrArray(strlen(upp->midiators));
	UINT i;
	for (i = 0; i < strlen(upp->midiators); i++)
	{
		char buf[6];
		sprintf(buf, "COM%c", upp->midiators[i]);
		(*midiator_array)[i] = new Midiator(buf, my_cdecl_read_hook, ewp);
	}

	inlh = get_input_nodes();	
	onlh = get_output_nodes();

   InputNodeList gni(inlh);
   OutputNodeList gno(onlh);

   gni.select_by_name( upp->default_input, &default_input_ioref, ewp);
	gno.select_by_name( upp->default_output, &default_output_ioref, ewp);
}

 MIDIsvc::
~MIDIsvc()
{
	closeAllInputs();
	closeAllOutputs();
	if (!CloseHandle(eventHandle))
		ewp->w("failed to close eventHandle in MIDIsvc cleanup");
   timeEndPeriod(1); // Windows call: clears a previously set minimum timer resolution. 
}


ioref
MIDIsvc::
open_input( const char* name, Bool allow_sysex )
{
	ioref n = default_input_ioref;
   InputNodeList gni(inlh);
	err_ind e = gni.select_by_name( name, &n, ewp);
	if (e)
	{
		ewp->w( "couldn't select an input node" ); return 0;
	}
	return open_input((ioref)n, allow_sysex);
}

ioref
MIDIsvc::
open_input(Bool allow_sysex)
{
	return open_input(default_input_ioref, allow_sysex);
}

void MIDIsvc::open_sysex(MyMidiIn *cn, ioref i)
{
		cn->sysex = new MySysexInputBuffer[my_sysex_num_buffers];
		cn->assembly_position = 0;
		int j;
		for (j = 0; j < my_sysex_num_buffers; j++)
		{
			MIDIHDR* hdr = &(cn->sysex[j].hdr);
			hdr->lpData = (char*)(cn->sysex[j].buf);
			hdr->dwBufferLength = my_sysex_input_size;
			hdr->dwFlags = 0;
			if (midiInPrepareHeader(cn->handle, hdr, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
			{
				ewp->w("on \"%s\" (node %d) unable to prepare sysex input buffer %d, will not try to "
					   "prepare or add any further sysex input buffers for this node", cn->szPname, i, j );
				break;
			}
			if (midiInAddBuffer(cn->handle, hdr, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
			{
				ewp->w("on \"%s\" (node %d) unable to add sysex input buffer %d, will not try to "
					   "prepare or add any further sysex input buffers for this node", cn->szPname, i, j );
				break;
			}
		}
}

ioref
MIDIsvc::
open_input( ioref i, Bool allow_sysex )
{
	MyMidiIn *cn = &((*inlh)[i]);
	if (cn->open) return i; // just return the number if already open

	// otherwise, we need to open it

   if (cn->ty == MyMidiIn::NORMAL)
	{
		MMRESULT mm = midiInOpen(&(cn->handle), i, (DWORD)my_read_hook, NULL, CALLBACK_FUNCTION);
		if (mm != MMSYSERR_NOERROR)
		{
			ewp->w( "couldn't open connection to \"%s\" (node %d)", cn->szPname, i ); return i;
		}

		if (allow_sysex) open_sysex(cn, i);
		else cn->sysex = NULL;

		mm = midiInStart(cn->handle);
		if (mm != MMSYSERR_NOERROR)
		{
			ewp->w( "couldn't start connection to \"%s\" (node %d)", cn->szPname, i ); return i;
		}
		midiInStart_time_offset = time();
	}
	else if (cn->ty == MyMidiIn::MIDIATOR)
	{
		cn->midiator->openInput(i-(cn->midiator->in_start));
	}
   else { ewp->e("unhandled node type"); throw rx(); }

	cn->open = True;
	return i;
}

ioref
MIDIsvc::
open_output( const char* name )
{
	ioref n = default_output_ioref;
   OutputNodeList gno(onlh);
	err_ind e = gno.select_by_name( name, &n, ewp);
	if (e)
	{
		ewp->w( "couldn't select an output node" ); return 0;
	}
	return open_output((ioref)n);
}

ioref
MIDIsvc::
open_output()
{
	return open_output(default_output_ioref);
}

ioref
MIDIsvc::
open_output( ioref i )
{
	ioref n = i;
	MyMidiOut *cn = &((*onlh)[n]);
	cn->running_status = 0; // clear out running status
	if (cn->open) // just return the number if already open
		return n;
	// otherwise, we need to open it

	if (cn->ty == MyMidiOut::NORMAL)
	{
		MMRESULT mm = midiOutOpen(&(cn->handle), n, (DWORD)my_read_hook, NULL, CALLBACK_FUNCTION);
		if (mm != MMSYSERR_NOERROR)
		{
			ewp->w( "couldn't open connection to \"%s\" (node %d)", cn->szPname, n ); return (ioref)n;
		}
	}
	else if (cn->ty == MyMidiOut::MIDIATOR)
	{
		cn->midiator->openOutput(n-(cn->midiator->out_start));
	}
	else { ewp->e("unhandled node type"); throw rx(); }

	cn->open = True; // remember that we opened it
	return (ioref)n;
}

void
MIDIsvc::
closeAllInputs()
{
	for (UINT i = 0; i < inlh->length; i++)
	{
		MyMidiIn* cn = &((*inlh)[i]);

		if (cn->open)
		{
			switch (cn->ty) {

			case MyMidiIn::MIDIATOR:
				cn->midiator->closeInput(i - (cn->midiator->in_start));
				cn->open = 0;
				break;

			case MyMidiIn::NORMAL:
				if (midiInReset(cn->handle) != MMSYSERR_NOERROR)
					ewp->w("on \"%s\" (node %d) midiInReset failed", cn->szPname, i);
				else if (cn->sysex != NULL)
				{
					int j;
					for (j = 0; j < my_sysex_num_buffers; j++)
					{
						MIDIHDR* hdr = &(cn->sysex[j].hdr);
						if (midiInUnprepareHeader(cn->handle, hdr, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
							ewp->w("on \"%s\" (node %d) unable to unprepare sysex input buffer", cn->szPname, i );
					}
					delete[](cn->sysex);
					cn->sysex = NULL;
				}
				if ((midiInStop(cn->handle) == MMSYSERR_NOERROR) && (midiInClose(cn->handle) == MMSYSERR_NOERROR))
					cn->open = 0;
				else
					ewp->w("unable to close \"%s\" (node %d)", cn->szPname, i);
            break;

         default:
     			ewp->e("unhandled node type"); throw rx();
			} // end switch
		} // end if
	} // end for
}

void
MIDIsvc::
closeAllOutputs()
{
	for (UINT i = 0; i < onlh->length; i++)
	{
		MyMidiOut* cn = &((*onlh)[i]);

		if (cn->open)
		{
			switch (cn->ty) {

         case MyMidiOut::MIDIATOR:
				cn->midiator->closeOutput(i - (cn->midiator->in_start));
				cn->open = 0;
				break;

			case MyMidiOut::NORMAL:
				if (midiOutClose(cn->handle) == MMSYSERR_NOERROR)
					cn->open = 0;
				else
					ewp->w("unable to close \"%s\" (node %d)", cn->szPname, i);
            break;

         default:
     			ewp->e("unhandled node type"); throw rx();
			} // end switch
		} // end if
	} // end for
}

ioref MIDIsvc::howManyOutputNodes() { return ioref(onlh->length); }

ioref MIDIsvc::howManyInputNodes()  { return ioref(inlh->length); }

const char*
MIDIsvc::
getOutputNodeName(ioref i)
{
	if (i >= onlh->length)
	{
		ewp->w("invalid output node %u",i);
		return "invalid";
	}
	return (*onlh)[i].szPname;
}

const char*
MIDIsvc::
getInputNodeName(ioref i)
{
	if (i >= inlh->length)
	{
		ewp->w("invalid input node %u",i);
		return "invalid";
	}
	return (*inlh)[i].szPname;
}

void
MIDIsvc::
my_cdecl_read_hook(HMIDIIN hMidiIn, UINT wMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	my_read_hook(hMidiIn, wMsg, dwInstance, dwParam1, dwParam2);
}

#pragma warning(disable:4100)
/*
to avoid warning C4100: 'dwInstance' : unreferenced formal parameter
*/

void CALLBACK
MIDIsvc::
my_read_hook(HMIDIIN hMidiIn, UINT wMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
	if (wMsg == MIM_DATA)
	{
		if (hMidiIn == timecode_input && timecode_input != NULL)
		{
			if ((dwParam1 & 0xf0ff) == 0xf1)
			{
				got_mtc();
			}
			else if ((dwParam1 & 0xf0ff) == 0x70f1)
			{
				switch ((dwParam1 >> 8) & 0xf)
				{
				case 0:
					fps = 24;
					break;
				case 1:
					fps = 25;
					break;
				default:
					fps = 30;
					break;
				}
			}
			if ((dwParam1 & 0xff) == 0xf1 && eat_timecode == True)
				return;
		}
		if (lp && handles(lp,dwParam1))
		{
			UINT n = 0;
			if (hMidiIn == NULL)
				n = (UINT)dwParam2; // Midiator passes its ioref in dwParam2
			else
			{
				for (UINT i = 0; i < inlh->length; i++)
					if ((*(inlh))[i].open && (*(inlh))[i].handle == hMidiIn)
						n = (UINT)i;
			}
			MyQueueElement q;
			q.dwParam1 = dwParam1;
			q.dwParam2 = dwParam2;
			q.n = n;
			q.hdr = NULL;
			cq.insertp(&q);
			schedp->processQueue(); // ask schedule thread to call our processQueue
		}
	}
	else if (wMsg == MIM_LONGDATA)
	{
		MIDIHDR* hdr = (MIDIHDR*) dwParam1;
		UINT n = 0;
		for (UINT i = 0; i < inlh->length; i++)
			if ((*(inlh))[i].open && (*(inlh))[i].handle == hMidiIn)
				n = (UINT)i;
		MyQueueElement q;
		q.hdr = hdr;
		q.n = n;
		cq.insertp(&q);
		schedp->processQueue();
	}
	else if (wMsg == MOM_DONE)
		SetEvent(eventHandle); // wake up write_ed(), waiting on sysex
	else if (wMsg == MIM_ERROR || wMsg == MIM_LONGERROR)
	{
		UINT n = 0;
		for (UINT i = 0; i < inlh->length; i++)
			if ((*(inlh))[i].open && (*(inlh))[i].handle == hMidiIn)
				n = (UINT)i;
		if (wMsg == MIM_ERROR)
			ewp->w("on \"%s\" (node %d) input error", (*(inlh))[n].szPname, n);
		else
			ewp->w("on \"%s\" (node %d) sysex input error", (*(inlh))[n].szPname, n);
	}
}

#pragma warning(default:4100)


void MIDIsvc::process_sysex_queue_el(MyQueueElement *q)
{
	MyMidiIn *cn = &((*inlh)[q->n]);

	cn->assembly_order[cn->assembly_position++] = q->hdr;
	if ((uchar)q->hdr->lpData[q->hdr->dwBytesRecorded-1] == 0xf7) // if end-of-sysex
	{
		int len = 0;
		int j;
		for (j = 0; j < cn->assembly_position; j++)
			len += cn->assembly_order[j]->dwBytesRecorded;
		uchar* tmp = new uchar[len];
		uchar* p = tmp;
		for (j = 0; j < cn->assembly_position; j++)
		{
			unsigned int k;
			for (k = 0; k < cn->assembly_order[j]->dwBytesRecorded; k++)
				*(p++) = cn->assembly_order[j]->lpData[k];
			if (midiInAddBuffer(cn->handle, cn->assembly_order[j], sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
				ewp->w("on \"%s\" (node %d) unable to add sysex input buffer", cn->szPname, q->n );
		}
		MIDI_msg m;
		m.node = ioref(q->n);
		m.set_sys_ex(tmp,len);
		delete[] tmp;
		if (lp && lp->msg_hnd[MidiMsg::system])
			lp->msg_hnd[MidiMsg::system](&m);
		cn->assembly_position = 0;
	}
}

void MIDIsvc::
processQueue()
{
	MyQueueElement *qe;
	int i = 0;
	while ( (qe = cq.headp()) != NULL )
	{
		if (qe->hdr != NULL) process_sysex_queue_el(qe);
		else
		{
			if (lp && handles(lp,qe->dwParam1)) // added this for safety 11/12/95 PP
         {
            MidiMsg m;
            WinMidiMsgToRog(qe->dwParam1,qe->dwParam2,qe->n,&m);
            lp->msg_hnd[m.msgt()](&m);
         }
		}
		cq.remove_head();
		i++;
	}
	if (i > 0 && upp->statistics_on) queueStat.throwInBin(i); // added for statistics gathering 11/12/95 PP
}

void
MIDIsvc::
write( const MIDI_msg *m )
{
#ifndef NDEBUG
	m->sanityCheck();
#endif
	if (m->ed != NULL) // handle extended message
	{
		if (m->ed->st == smsgt::file_meta)
			return; // ignore meta messages
		csp->enter(); // write_ed is not reentrant
		write_ed(m);
		csp->leave();
		return;
	}
	DWORD p;
	UINT n;
   RogMidiMsgToWin(*m, &p, &n);
	if (n >= (UINT)(onlh->length))
	{
		ewp->w("invalid output node %u",n);
		return;
	}

	MyMidiOut* cn = &((*onlh)[n]);

	if (cn->open == False)
	{
		ewp->w("on \"%s\" (node %d) attempt to write output when not open",cn->szPname,n);
		return;
	}

	if (cn->ty == MyMidiOut::NORMAL)
	{
		csp->enter();

		if (upp->running_status_on)
		{
			if ((p & 0xf0) == 0xf0 || cn->running_status_time > prev_time) // don't use running status for system messages; plus always throw in status byte every half second
			{
				cn->running_status = 0;
				cn->running_status_time = prev_time + 1000;
			}
			if ((p & 0xff) == cn->running_status)
				p = p >> 8; // get rid of status byte if not needed
			else
				cn->running_status = (uchar)(p & 0xff); // keep track of running status
		}

		MMRESULT mm = midiOutShortMsg(cn->handle, p);
		if (mm != MMSYSERR_NOERROR)
		{
			char errtext[MAXERRORLENGTH+1];
			midiOutGetErrorText(mm, errtext, MAXERRORLENGTH);
			ewp->w("on \"%s\" (node %d) output error -- %s",cn->szPname,n,errtext);
		}
		csp->leave();
	}
	else if (cn->ty == MyMidiOut::MIDIATOR)
	{
		csp->enter();
		cn->midiator->send(n - cn->midiator->out_start, p);
		csp->leave();
	}
	else { ewp->e("unhandled node type"); throw rx(); }
}

void
MIDIsvc::
write_ed(const MIDI_msg *m)
{
	if (m->ed->st != smsgt::sys_ex)
	{
		ewp->w("attempted to write extended message with unrecognized ed type");
		return;
	}
	UINT n = (UINT)m->node;
	if (n >= (UINT)(onlh->length))
	{
		ewp->w("invalid output node %u",n);
		return;
	}

	MyMidiOut* cn = &((*onlh)[n]);

	if (!cn->open)
	{
		ewp->w("on \"%s\" (node %d) attempt to write output when not open",cn->szPname,n);
		return;
	}

	if (cn->ty == MyMidiOut::NORMAL)
	{
		cn->running_status = 0; // zap running status so we don't get confused later
		MIDIHDR mh;
		mh.lpData = (LPSTR)(m->ed->sys_ex.d);
		mh.dwBufferLength = m->ed->sys_ex.l;
		mh.dwFlags = 0;
		if (midiOutPrepareHeader(cn->handle, &mh, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
		{
			ewp->w("on \"%s\" (node %d) unable to prepare MIDIHDR for sysex",cn->szPname,n);
			return;
		}
		if (midiOutLongMsg(cn->handle, &mh, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
		{
			ewp->w("on \"%s\" (node %d) unable to send sysex",cn->szPname,n);
			midiOutUnprepareHeader(cn->handle, &mh, sizeof(MIDIHDR)); // try to clean up
			return;
		}
		DWORD result = WaitForSingleObject(eventHandle, 1000); // wait for sysex to finish
		if (result == WAIT_TIMEOUT)
			ewp->w("on \"%s\" (node %d) timed out sending sysex",cn->szPname,n);
		if (midiOutUnprepareHeader(cn->handle, &mh, sizeof(MIDIHDR)) != MMSYSERR_NOERROR)
		{
			ewp->w("on \"%s\" (node %d) unable to unprepare MIDIHDR for sysex",cn->szPname,n);
			return;
		}
	}
	else if (cn->ty == MyMidiOut::MIDIATOR)
	{
		ewp->w("on \"%s\" (node %d) system exclusive not supported yet for MIDIator output",cn->szPname,n);
		return;
	}
	else { ewp->e("unhandled node type"); throw rx(); }
}

void MIDIsvc::RogMidiMsgToWin( const MidiMsg& m, DWORD* p, UINT* n )
{
	*p = m.s + (m.d1<<8) + (m.d2 << 16);
	*n = m.node;
}

void MIDIsvc::WinMidiMsgToRog( DWORD p1, DWORD p2, UINT n, MidiMsg *m )
{
	m->time = (p2<<1) + get_adj_midiInStart_time_offset(); // used to be time = oap->time();
	m->node = (ioref) n;
	m->s  = uchar((p1      ) & 0xff);
	m->d1 = uchar((p1 >>  8) & 0xff);
	m->d2 = uchar((p1 >> 16) & 0xff);
}

//--------------------------------------------------------------------------------------------------

#include "string.h" // for strstr()

err_ind
MIDIsvc::GenericNodeList::
select_by_name( const char *in_name, ioref *id, err_and_warn *ewp )
{
	UINT num_nodes = length();

	char in_name_upper[MAXPNAMELEN+1];
	strtoupper(in_name, in_name_upper);

	int none_found = 1;
	
	for (ioref i = 0; i < num_nodes && none_found; i++)
	{
		char tmp[MAXPNAMELEN+1];
		strtoupper( name(i), tmp);
				
		if (strstr(tmp, in_name_upper))
		{
			ewp->i( "SUCCESS: \"%s\" matched to \"%s\"", in_name, name(i) );
			*id = i;
			none_found = 0;
		}
	}
	if (none_found)
		if (num_nodes > 0)
		{	
			ewp->w( "Couldn't find \"%s\", using \"%s\" instead", in_name, name(*id) );
		}
		else { ewp->w("No input nodes exist"); return err; }
		
	return no_err;
}

// Authors: Ben Denckla, Herb Yang, Patrick Pelletier
