#include "NotePairScore.h"
#include "TrackMerger.h"

static void convertTime(MidiMsgArrayPtr p, long division)
{
	if (p->length < 1)
		return;
	// change tick time into real time
	ScoreTime lastTime = 0;
	long lastTick = 0;
	long currentTempo = 500000;

	for (UINT i = 0; i < p->length; i++)
	{
		MIDI_msg* m = &((*p)[i]);
		long thisTick = m->time;
		double multiplier = double(currentTempo)/double(division)/double(500);
		
		m->time = (long) (lastTime + (thisTick-lastTick)*multiplier + 0.5);
		
		if (m->is_meta(meta::tempo))
		{
			long newTempo = m->ed->meta_event.tem;
			lastTime += (ScoreTime) ((thisTick-lastTick)*multiplier + 0.5);
			lastTick = thisTick;
			currentTempo = newTempo;
		}
	}
}

int NotePairScore::dual_purpose_thingy(const ScoreCreationInfo* sci_const, NotePairScoreElementArray* npsea)
{
	ScoreCreationInfo* sci = (ScoreCreationInfo*) sci_const; // cast away constness because we maintain logical constness
	sci->do_sort();
	UINT i;
	UINT numtracks = 0;
	for (i = 0; i < sci->track_included.length; i++)
		if (sci->track_included[i])
			numtracks++;
	if (sci->num_msgs > 0)
		numtracks++;
	int2 division = sci->division;
	if (numtracks < 1)
	{
		return 0;
	}
	MidiMsgArrayPtrArrayPtr p = new MidiMsgArrayPtrArray(numtracks);
	UINT i2 = 0;
	for (i = 0; i < sci->track_included.length; i++)
	{
		if (!sci->track_included[i])
			continue;
		ioref node = sci->track_node[i];
		MidiMsgArrayPtr p1 = sci->fd->getTrack(i);
		MidiMsgArrayPtr p2 = new MidiMsgArray(p1->length);
		if (p1->length < 1)
			continue;
		UINT j;
		for (j = 0; j < p1->length; j++) // copy MidiMsgs to MIDI_msgs
		{
			MidiMsg* m = &((*p1)[j]);
			m->sanityCheck();
			*((MidiMsg*)&(*p2)[j]) = *m;
			(*p2)[j].node = node;
		}
		(*p)[i2++] = p2;
	}
	if (sci->num_msgs > 0) // handle manually inserted MIDI_msgs
	{
		MidiMsgArrayPtr p2 = new MidiMsgArray(sci->num_msgs);
		UINT j;
		for (j = 0; j < p2->length; j++)
			(*p2)[j] = *(sci->msgs[j]);
		(*p)[i2++] = p2;
	}
	if (i2 != numtracks)
		sci->r->ew.w("Sanity check failure in NotePairScore::dual_purpose_thingy");
	MidiMsgArrayPtr msgs = TrackMerger<MIDI_msg>::merge(p); // merge all the tracks together
	convertTime(msgs, division);
	MIDI_msg dummy;
	const char* dummy_msg = "zap me";
	dummy.set_meta(meta::txt_generic, (const uchar*)dummy_msg, strlen(dummy_msg));
	i2 = 0;
	// pair up the note-ons and note-ends
	for (i = 0; i < msgs->length; i++)
	{
		MIDI_msg* m = &((*msgs)[i]);
		if (m->is_ext() && m->ed->st == smsgt::file_meta)
			continue; // leave out meta messages
		if (m->msgt() == MidiMsg::note_on && m->vel() != 0)
		{
			UINT j;
			for (j = i; j < msgs->length; j++)
			{
				MIDI_msg* mm = &((*msgs)[j]);
				if (m->is_on_for(*mm))
				{
					NotePairScoreTime dur = mm->time - m->time;
					*mm = dummy;
					if (npsea != NULL)
					{
						(*npsea)[i2].msg = *m;
						(*npsea)[i2].dur = dur;
					}
					i2++;
					break;
				}
			}
			if (j < msgs->length)
				continue;
			// no off found
			sci->r->ew.w("Note-on found with no matching note-end");
			if (npsea != NULL)
			{
				(*npsea)[i2].msg = *m;
				(*npsea)[i2].dur = -1;
			}
			i2++;
		}
		else if (m->msgt() == MidiMsg::control_chn && m->d1 == 64 && m->d2 == 0x7f) // sustain pedal down
		{
			UINT j;
			for (j = i; j < msgs->length; j++)
			{
				MIDI_msg* mm = &((*msgs)[j]);
				if (mm->msgt() == MidiMsg::control_chn && mm->d1 == 64 && mm->d2 == 0x00) // sustain pedal up
				{
					NotePairScoreTime dur = mm->time - m->time;
					*mm = dummy;
					if (npsea != NULL)
					{
						(*npsea)[i2].msg = *m;
						(*npsea)[i2].dur = dur;
					}
					i2++;
					break;
				}
			}
			if (j < msgs->length)
				continue;
			// no off found
			sci->r->ew.w("Sustain-pedal-down found with no matching sustain-pedal-up");
			if (npsea != NULL)
			{
				(*npsea)[i2].msg = *m;
				(*npsea)[i2].dur = -1;
			}
			i2++;
		}
		else // not a paired MIDI_msg
		{
			if (npsea != NULL)
			{
				(*npsea)[i2].msg = *m;
				(*npsea)[i2].dur = -1;
			}
			i2++;
		}
	} // end of for loop
	if (npsea != NULL && i2 != npsea->length)
		sci->r->ew.w("sanity check failure in NotePairScore::dual_purpose_thingy");
	return i2;
}

#pragma warning(disable : 4097)
/*
warning C4097: typedef-name 'NotePairScoreElementArray' used as 
synonym for class-name 'UtilArray<class NotePairScoreElement>'
*/

NotePairScore::NotePairScore(const ScoreCreationInfo* sci) :
   NotePairScoreElementArray(dual_purpose_thingy(sci, NULL))
{
	dual_purpose_thingy(sci, this);
}

#pragma warning(default : 4097)

// Author: Patrick Pelletier
