/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "State.h"

ChannelState State::defaultChannelState;

static void iterator1(ioref n, int c, Bool* e, void* x)
{  // create the ChannelStates where map has 1
	State* that = (State*) x;
	if (*e == 1)
		(*that)(n, c) = new ChannelState();
}

State::State(Score::ChannelUseMapPtr u, ioref non) : PerChannel<ChannelStatePtr>(non)
{
	u->iterate(iterator1, (void*) this); // create the ChannelStates where map has 1
}

void State::update(const MIDI_msg* m)
{
	ioref n = m->node;
	if (m->msgt() == MidiMsg::system)
		return; // don't do non-channel messages
	int c = m->chan();
	ChannelState* cs = operator()(n, c);  // look up the right ChannelState
	if (cs != NULL) // make sure there is a ChannelState
		cs->update(m); // pass along the message
}


static void iterator2(ioref n, int c, ChannelStatePtr* e, void* x)
{
   State::my_struct* temp = (State::my_struct*) x;
	State* s = (State*) temp->s; // we can cast away constness because we don't really want to modify it
	ChannelState* cs1 = *e; // the ChannelState for our State
	ChannelState* cs2 = (*s)(n, c); // the ChannelState for the other State
	if (cs1 == NULL && cs2 == NULL)
		return; // nothing more to do if no state for either channel
	if (cs1 == NULL)
		cs1 = temp->dcs;;
	if (cs2 == NULL)
		cs2 = temp->dcs;
	cs1->change(cs2, n, c, temp->outfunc);
}

void State::change(const State* s, void (*outfunc)(const MIDI_msg&))
{
	my_struct temp = {s, outfunc, &defaultChannelState};
	iterate(iterator2, (void*) (&temp)); // iterate over all the ChannelStates
}

// Author: Patrick Pelletier
