/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#ifndef NOTEPAIRSCOREPLAYER_H
#define NOTEPAIRSCOREPLAYER_H

#include "NotePairScoreElement.h"
#include "NotePairScore.h"
#include "Rogus_McBogus.h"
#include "WrappedCS.h"

class NotePairScorePlayer {
public:
	typedef UtilPtr<NotePairScore> NotePairScorePtr;
	typedef UtilPtr<NotePairScorePlayer> NotePairScorePlayerPtr;
	typedef UtilArray<NotePairScorePlayerPtr> NotePairScorePlayerPtrArray;
   typedef void (*ofp)(const NotePairScoreElement&, NotePairScoreScale, ScheduleTime);
private:
	NotePairScorePtr score;
	static ScheduleTime callback(void* dwUser); 
	Bool playing;
	Bool muted;
	Bool looped;
	volatile task* ott;
	NotePairScoreScale scale;
	int index;
	NotePairScoreTime leftover;
   ofp outfunc;
   //#void (*outfunc)(const NotePairScoreElement&, NotePairScoreScale, ScheduleTime);
	ScheduleTime lastWakeUp;
	ScheduleTime nextWakeUp;
	static void defaultOutFunc(const NotePairScoreElement&, NotePairScoreScale, ScheduleTime);

	void do_start(ScheduleTime);
	err_ind do_stop(ScheduleTime);

   static Rogus_McBogus *r;
   static WrappedCS cs;

public:
	static void startMultiple(NotePairScorePlayerPtrArray*);
	static void stopMultiple(NotePairScorePlayerPtrArray*);
	NotePairScorePlayer
   (
      Rogus_McBogus*,
      NotePairScorePtr,
      ofp func = defaultOutFunc
   );
	~NotePairScorePlayer();
	void start();
	void stop();
	Bool isPlaying() const { return playing; }
	void setMute(Bool);
	Bool getMute() const { return muted; }
	void setLoop(Bool l) { looped = (l ? 1 : 0); }
	Bool getLoop() const { return looped; }
	void setIndex(int);
	void rewind() { setIndex(0); }
	int getIndex() const { return index; }
	void setTimeUnscaled(ScoreTime);
	NotePairScoreTime getTimeUnscaled() const;
	void setTimeScaled(NotePairScoreTime t) { setTimeUnscaled((NotePairScoreTime)(t / scale)); }
	NotePairScoreTime getTimeScaled() const { return (NotePairScoreTime) (getTimeUnscaled() * scale); }
	void setScale(NotePairScoreScale);
	NotePairScoreScale getScale() const { return scale; }
};

#endif
// Author: Patrick Pelletier
