/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#ifndef SCOREPLAYER_H
#define SCOREPLAYER_H

#include "UtilPtr.h"
#include "Score.h"
#include "State.h"
#include "MidiFilter.h"
#include "Rogus_McBogus.h"
#include "WrappedCS.h"

typedef float ScoreScale;

class ScorePlayer {
public:
	typedef UtilPtr<Score> ScorePtr;
	typedef UtilPtr<MidiFilter> MidiFilterPtr;
	typedef UtilPtr<ScorePlayer> ScorePlayerPtr;
	typedef UtilArray<ScorePlayerPtr> ScorePlayerPtrArray;
private:
	typedef UtilPtr<State> StatePtr;
	ScorePtr score;
	Score::ChannelUseMapPtr u;
	StatePtr state;
	static ScheduleTime callback(void* dwUser); 
	Bool playing;
	Bool muted;
	Bool looped;
	volatile task* ott;
	ScoreScale scale;
	int index;
	static StatePtr defaultState;
	ScoreTime leftover;
	void (*outfunc)(const MIDI_msg&);
	ScheduleTime lastWakeUp;
	ScheduleTime nextWakeUp;
	void do_start(ScheduleTime);
	err_ind do_stop(ScheduleTime);
	static StatePtr calculate_state(Score::ChannelUseMapPtr, ScorePtr, int);
	typedef UtilArray<MidiFilterPtr> MidiFilterPtrArray;
	typedef UtilPtr<MidiFilterPtrArray> MidiFilterPtrArrayPtr;
	MidiFilterPtr dummy; // not used, but won't compile without this
	MidiFilterPtrArrayPtr filters;
	int find_filter(const char*) const;
	static void defaultOutFunc(const MIDI_msg&);
	task* functask;
	int funcnum;
	static ScheduleTime func_callback(void* v);
	Bool calling_func;
	ScoreTime decide_next_function();
	ScoreTime decide_next_function_for_callback();
   static Rogus_McBogus *r;
   static WrappedCS cs;
public:
	static void startMultiple(ScorePlayerPtrArray*);
	static void stopMultiple(ScorePlayerPtrArray*);
	ScorePlayer(Rogus_McBogus*, ScorePtr, void (*func)(const MIDI_msg&) = defaultOutFunc);
	~ScorePlayer();
	void start();
	void stop();
	Bool isPlaying() const { return playing; }
	void setMute(Bool);
	Bool getMute() const { return muted; }
	void setLoop(Bool l) { looped = (l ? 1 : 0); }
	Bool getLoop() const { return looped; }
	void setIndex(int);
	void rewind() { setIndex(0); }
	int getIndex() const { return index; }
	void setTimeUnscaled(ScoreTime);
	ScoreTime getTimeUnscaled() const;
	void setTimeScaled(ScoreTime t) { setTimeUnscaled((ScoreTime)(t / scale)); }
	ScoreTime getTimeScaled() const { return (ScoreTime) (getTimeUnscaled() * scale); }
	void setScale(ScoreScale);
	ScoreScale getScale() const { return scale; }
	void addFilter(MidiFilterPtr);
	void removeFilter(const char*);
	MidiFilterPtr getFilter(const char*) const;
};

#endif
// Author: Patrick Pelletier
